7#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# grabslides : multipurpose frame grabber
#
# (c) 2005-2009 Jerome Alet <alet@librelogiciel.com>
# (c) 2009 University of New-Caledonia
# (c) 2005, 2006 Faculty of Medecine, University of Nice-Sophia Antipolis
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
#

"""This module defines the base class used by all kinds of frame grabbers."""

import sys
import os
import time
import errno
import signal
import threading
import tempfile
import Queue

from PIL import Image

import gobject

import pygst
pygst.require("0.10")
import gst

EVERY = 5 # Minimal number of seconds between slides.
TIMEOUT = 5 # Maximal time to wait for the pipeline to stop

class BaseFrameGrabber :
    """A class to grab frames from video streams."""
    def __init__(self, parent, url, every, threshold, source=""):
        """Initializes the frame grabber."""
        self.parent = parent
        self.url = url
        self.every = every
        self.threshold = threshold
        self.shuttingDown = False
        self.mainloop = None
        self.pipeline = None
        self.watchBusMessages = None
        self.seenEOS = False
        self.grabit = False
        self.buffer_probes = {}
        self.width = None
        self.height = None
        self.begintime = None
        self.endtime = None
        self.definition = " ! ".join([ source,
                                       "decodebin",
                                       "ffmpegcolorspace",
                                       "video/x-raw-rgb,depth=24,bpp=24",
                                       'fakesink name="Frame Grabber" signal-handoffs=True',
                                     ])
        self.framequeue = Queue.Queue(0)
        self.quitEvent = threading.Event()

    def diffImages(self, image1, image2) :
        """ Compare 2 images, returns True if their difference is higher than threshold, else False."""
        if (not image1) or (not image2) :
            return True
        image1 = image1.getdata()
        image2 = image2.getdata()
        nbpix1 = len(image1)
        nbpix2 = len(image2)
        if nbpix1 != nbpix2 :
            return True
        threshold = self.threshold      # small speedup
        increment = 100.0 / nbpix1
        pixdiff = 0.0
        for i in range(nbpix1) :
            if abs(sum(image1[i]) - sum(image2[i])) > 50.0 :
                pixdiff += increment
                if pixdiff >= threshold :
                    return True
        self.parent.logDebug("Diff : %.3f%%  <  Threshold : %.3f%%" % (pixdiff, threshold))
        return False

    def handleMessages(self, bus, message) :
        """Handles messages on the bus."""
        if message.type == gst.MESSAGE_EOS :
            self.parent.logDebug("++++++++++++ EOS +++++++++++++")
            self.stop()

    def run(self) :
        """Starts a MainLoop for this pipeline."""
        self.parent.logDebug("Starting Main Loop...")
        self.mainloop = gobject.MainLoop()
        threading.Thread(target=self.mainloop.run).start()
        self.parent.logDebug("Main Loop started.")

    def shutdown(self) :
        """Shuts the MainLoop down."""
        if not self.shuttingDown :
            self.shuttingDown = True
            self.stop()
            if self.mainloop is not None :
                self.parent.logDebug("Shutting down Main Loop...")
                gobject.idle_add(self.mainloop.quit)

    def stop(self) :
        """Removes the buffer probe and stops the pipeline."""
        if self.pipeline is not None :
            for sink in self.pipeline.sinks() :
                name = sink.get_name()
                factoryname = sink.get_factory().get_name()
                if factoryname == "fakesink" :
                    pad = sink.get_pad("sink")
                    pad.remove_buffer_probe(self.buffer_probes[name])
                    del self.buffer_probes[name]
                    self.parent.logDebug("Buffer probe removed from %(name)s" % locals())
            self.pipeline.set_state(gst.STATE_NULL)
            self.pipeline = None
            self.quitEvent.set()
            self.endtime = time.time()
            self.parent.logDebug("Pipeline stopped.")

    def start(self) :
        """Installs the buffer probe and starts the pipeline."""
        self.seenEOS = False
        self.pipeline = gst.parse_launch(self.definition)
        self.parent.logDebug("Pipeline created.")
        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        self.watchBusMessages = bus.connect("message", self.handleMessages)
        self.parent.logDebug("Messages handler activated.")
        for sink in self.pipeline.sinks() :
            name = sink.get_name()
            factoryname = sink.get_factory().get_name()
            if factoryname == "fakesink" :
                pad = sink.get_pad("sink")
                self.buffer_probes[name] = pad.add_buffer_probe(self.buffer_probe_handler, name)
                self.parent.logDebug("Buffer probe added for %(name)s" % locals())
        self.begintime = time.time()
        self.pipeline.set_state(gst.STATE_PLAYING)
        self.parent.logDebug("Pipeline started.")

    def buffer_probe_handler(self, pad, buffer, name) :
        """Capture buffers as PIL Images when told to."""
        if self.grabit :
            caps = buffer.caps
            if caps is not None :
                filters = caps[0]
                self.width = filters["width"]
                self.height = filters["height"]
            try :
                timecode = self.pipeline.query_position(gst.FORMAT_TIME)[0] / gst.SECOND
            except gst.QueryError :
                timecode = time.time() - self.begintime
            frame = Image.frombuffer("RGB",
                                    (self.width, self.height),
                                    buffer.data,
                                    "raw",
                                    "RGB",
                                    0,
                                    1)
            self.framequeue.put((timecode, frame), True)
            self.grabit = False
        else :
            pass
        return True

    def seekToNextFrame(self, interval) :
        """Seeks stream to next position."""
        if self.pipeline is None :
            return False
        else :
            time.sleep(interval)
            return True

    def grabFrame(self) :
        """Returns the first frame in the queue."""
        try :
            return self.framequeue.get(False)
        except Queue.Empty :
            return (None, None)

    def grab(self) :
        """Grabs all wanted frames."""
        self.start()
        nextok = True
        previmage = None
        while nextok and not self.quitEvent.isSet() :
            self.grabit = True
            (tc, curimage) = self.grabFrame()
            if tc and curimage :
                if self.diffImages(previmage, curimage) :
                    previmage = curimage.copy()
                    self.parent.saveSlide(curimage, tc)
                nextok = self.seekToNextFrame(self.every)
